package edu.berkeley.sbp;
import edu.berkeley.sbp.*;
import edu.berkeley.sbp.util.*;
import edu.berkeley.sbp.Parser.Table.*;
import edu.berkeley.sbp.Sequence.Position;
import java.io.*;
import java.util.*;
import java.lang.reflect.*;

/** implements Tomita's Graph Structured Stack */
class GSS {

    public GSS() { }

    private Phase.Node[] reducing_list = null;
    public int resets = 0;
    public int waits = 0;

    HashMapBag<Integer,Sequence>       inhibited       = new HashMapBag<Integer,Sequence>();
    HashMapBag<Integer,Sequence>       expectedInhibit = new HashMapBag<Integer,Sequence>();
    HashMapBag<Sequence,Phase.Waiting> waiting         = new HashMapBag<Sequence,Phase.Waiting>();
    HashMapBag<Integer,Sequence>       performed       = new HashMapBag<Integer,Sequence>();
    
    /** FIXME */
    public  Forest.Ref finalResult;

    /** corresponds to a positions <i>between tokens</i> the input stream; same as Tomita's U_i's */
    public class Phase<Tok> implements Invokable<State, Forest, Phase<Tok>.Node>, IntegerMappable {

        public void invoke(State st, Forest result, Node n) {
            good |= next.newNode(n, result, st, false);
        }

        /** the token immediately after this phase */
        final Tok token;

        private final int pos;

        boolean reducing;
        private IntPairMap<Phase.Node> hash;  /* ALLOC */
        private IntPairMap<Forest> singularReductions;  /* ALLOC */
        private boolean closed;
        private boolean good;
        private Phase next = null;
        private Phase prev;
        private Input.Location location;
        public final Parser parser;

        private Forest forest;

        public Phase(Phase prev, Parser parser, Phase previous, Tok token, Input.Location location, Forest forest) {
            this.prev = prev;
            this.forest = forest;
            this.parser = parser;
            this.pos = previous==null ? 0 : previous.pos+1;
            this.token = token;
            this.location = location;
            inhibited.clear();
            reset();
        }

        public void reset() {
            waiting.clear();
            performed.clear();
            hash = new IntPairMap<Phase.Node>();
            singularReductions = new IntPairMap<Forest>();
            expectedInhibit.clear();
            expectedInhibit.addAll(inhibited);
            good = false;
            closed = false;
            reducing = false;
            finalResult = null;
            if (prev != null) prev.shift(this, forest);
        }

      
        public boolean isDone() throws ParseFailed {
            if (token != null) return false;
            if (token==null && finalResult==null)
                throw new ParseFailed(ParseFailed.error(ANSI.red("unexpected end of file\n"), token, hash.values()), getLocation());
            return true;
        }

        public Input.Location getLocation() { return location; }

        /** add a new node (merging with existing nodes if possible)
         *  @param parent             the parent of the new node
         *  @param result             the SPPF result corresponding to the new node
         *  @param state              the state that the new node is in
         *  @param fromEmptyReduction true iff this node is being created as a result of a reduction of length zero (see GRMLR paper)
         *  @param start              the earliest part of the input contributing to this node (used to make merging decisions)
         */
        public boolean newNode(Node parent, Forest pending, State state, boolean fromEmptyReduction) {
            Node p = hash.get(state, parent==null?null:parent.phase());
            if (p != null)  return newNode2(p, parent, pending, state, fromEmptyReduction);
            else            return newNode3(parent, pending, state, fromEmptyReduction);
        }
        public void newNode(Node parent, Forest pending, State state, boolean fromEmptyReduction, Position reduction) {
            int pos = parent==null?0:parent.phase()==null?0:parent.phase().pos;
            Sequence owner = reduction==null ? null : reduction.owner();
            if (reduction!=null) {
                if (inhibited.contains(pos, owner)) return;
                if (owner.needs != null)
                    for(Sequence s : owner.needs)
                        if (!performed.contains(pos, s)) {
                            waiting.add(s, new Waiting(parent, pending, state, fromEmptyReduction, reduction));
                            return;
                        }
                if ((owner.needed != null && owner.needed.size()>0) ||
                    (owner.hated != null && owner.hated.size()>0) ||
                    (owner.hates != null && owner.hates.size()>0))
                    performed.add(pos, owner);
            }
            if (!owner.lame)
                newNode(parent, pending, state, fromEmptyReduction);
            if (reduction!=null) inhibit(reduction, parent==null?0:parent.phase().pos);
            if (reduction != null) {
                boolean redo = true;
                while(redo) {
                    redo = false;
                    for(Waiting w : waiting.getAll(owner)) {
                        if (w.parent==parent || (parent!=null&&w.parent!=null&&w.parent.phase()==parent.phase())) {
                            waiting.remove(owner, w);
                            w.perform();
                            redo = true;
                            break;
                        }
                    }
                }
            }
        }
        private boolean newNode2(Node p, Node parent, Forest pending, State state, boolean fromEmptyReduction) {
            p.holder.merge(pending);
            if (p.parents().contains(parent)) return true;
            p.parents().add(parent, true);
            if (p!=parent && !fromEmptyReduction && reducing) p.performReductions(parent);
            return true;
        }
        private boolean newNode3(Node parent, Forest pending, State state, boolean fromEmptyReduction) {
            do {
                if (token != null && state.canShift(token)) break;
                if (state.isAccepting()) break;
                if (token==null) break;
                if (!state.canReduce(token)) return false;
                //if (count > 1) break;
                //if (r.numPop == 0) break;
                //r.reduce(pending, parent, null, Phase.this, null);
                //return;
            } while(false);

            Node n = new Node(parent, pending, state);  // ALLOC
            if (reducing) {
                n.performEmptyReductions();
                if (!fromEmptyReduction) n.performReductions(parent);
            }
            return true;
        }

        public void uninhibit(int p, Sequence s) {
            if (s.hated!=null)
                for(Sequence s2 : s.hated)
                    inhibited.remove(p, s2);
        }

        public void inhibit(Position r, int p) {
            if (r.owner().hated == null) return;
            // remember that dead states are still allowed to shift -- just not allowed to reduce
            boolean reset = false;
            for(Sequence seq : r.owner().hated) {
                if (performed.contains(p,seq)) {
                    uninhibit(p, seq);
                    //System.out.println("\nresetting due to " + r.owner() + " killing " + seq);
                    //inhibited.clear();
                    inhibited.add(p, seq);
                    //inhibited = new HashMapBag<Integer,Sequence>();
                    reset = true;
                    resets++;
                    throw new Reset();
                }
                inhibited.add(p, seq);
                expectedInhibit.remove(p, seq);
            }
        }
        
        /** perform all reduction operations */
        public void reduce() {
            try {
                reducing = true;
                if (reducing_list==null || reducing_list.length < hash.size())
                    reducing_list = new Phase.Node[hash.size() * 4];
                hash.toArray(reducing_list);
                int num = hash.size();
                for(int i=0; i<num; i++) {
                    Node n = reducing_list[i];
                    n.performEmptyReductions();
                    // INVARIANT: we never "see" a node until its parent-set is complete, modulo merges
                }
                for(int i=0; i<num; i++) {
                    Node n = reducing_list[i];
                    reducing_list[i] = null;
                    n.performReductions();
                }
                if (expectedInhibit.size() > 0) {
                    inhibited.removeAll(expectedInhibit);
                    System.out.println("\n!!!!\n");
                    throw new Reset();
                }
            } catch (Reset r) {
                reset();
                reduce();
            }
        }

        class Reset extends RuntimeException { }

        /** perform all shift operations, adding promoted nodes to <tt>next</tt> */
        public void shift(Phase next, Forest result) throws ParseFailed {
            // this massively improves GC performance
            if (prev!=null) {
                prev.hash = null;
                prev.singularReductions = null;
            }
            this.next = next;
            closed = true;
            Forest res = null;
            boolean ok = false;
            for(Phase.Node n : hash.values()) {
                if (token == null && n.state.isAccepting()) {
                    if (finalResult==null) finalResult = new Forest.Ref();
                    finalResult.merge(n.holder);
                }
                if (token == null) continue;
                n.state.invokeShifts(token, this, result, n);
            }

            if (!good && token!=null)
                throw new ParseFailed(ParseFailed.error(ANSI.red("unexpected character ")+" \'"+
                                                        ANSI.purple(StringUtil.escapify(token+"", "\\\'\r\n"))+
                                                        "\' encountered at "+
                                                        ANSI.green(getLocation())+"\n", token, hash.values()),
                                        getLocation());
            if (token==null && finalResult==null)
                throw new ParseFailed(ParseFailed.error(ANSI.red("unexpected end of file\n"), token, hash.values()),
                                        getLocation());
        }


        public class Waiting {
            Node parent;
            Forest pending;
            State state;
            boolean fromEmptyReduction;
            Position reduction;
            public Waiting(Node parent, Forest pending, State state, boolean fromEmptyReduction, Position reduction) {
                waits++;
                this.parent = parent;
                this.pending = pending;
                this.state = state;
                this.fromEmptyReduction = fromEmptyReduction;
                this.reduction = reduction;
            }
            public void perform() {
                //System.out.println("performing: " + reduction.position);
                newNode(parent, pending, state, fromEmptyReduction, reduction);
            }
        }
       
        // Node /////////////////////////////////////////////////////////////////////////////////

        /** a node in the GSS */
        public final class Node extends FastSet<Node> implements Invokable<Position, Node, Node>, IntegerMappable {

            private Forest.Ref holder = null;
            private boolean allqueued = false;

            /** what state this node is in */
            public final Parser.Table<Tok>.State<Tok> state;

            /** which Phase this Node belongs to (node that Node is also a non-static inner class of Phase) */
            public  Phase phase() { return Phase.this; }
            public  Forest.Ref holder() { return holder==null ? (holder = new Forest.Ref()) : holder; }
            public  Forest pending() { return Phase.this.closed ? holder().resolve() : holder; }
            public  FastSet<Node> parents() { return this; }

            public void performReductions() {
                if (allqueued) return;
                allqueued = true;
                state.invokeReductions(token, this, this, null);
            }

            public void performReductions(Node n2) {
                if (!allqueued) performReductions();
                else            state.invokeReductions(token, this, this, n2);
            }

            public final void invoke(Position r, Node n, Node n2) {
                if (n==null || n2==null || r.pos==0) {
                    if (r.pos==0) {
                        if (n==null) n = this;
                        else return;
                    }
                    if (n==null) return;
                    Forest[] holder = new Forest[r.pos];
                    if (r.pos==0) n.finish(r, r.zero(), n.phase(), holder);
                    else                   n.reduce(r, r.pos-1, n.phase(), holder);
                } else {
                    Forest[] holder = new Forest[r.pos];
                    if (r.pos<=0) throw new Error("called wrong form of reduce()");
                    int pos = r.pos-1;
                    Forest old = holder[pos];
                    holder[pos] = n.pending();
                    if (pos==0) {
                        System.arraycopy(holder, 0, r.holder, 0, holder.length);
                        Forest rex = null;
                        if (r.pos==1)  rex = singularReductions.get(this, r);
                        if (rex==null) {
                            rex = r.rewrite(n.phase().getLocation());
                            if (r.pos==1) singularReductions.put(this, r, rex);
                        }
                        n2.finish(r, rex, n.phase(), holder);
                    } else {
                        n2.reduce(r, pos-1, n.phase(), holder);
                    }
                    holder[pos] = old;
                }
            }

            public void reduce(Position r, int pos, Phase target, Forest[] holder) {
                Forest old = holder[pos];
                holder[pos] = this.pending();
                if (pos==0) {
                    System.arraycopy(holder, 0, r.holder, 0, holder.length);
                    for(int i=0; i<r.pos; i++) if (r.holder[i]==null) throw new Error("realbad");
                    Forest rex = null;
                    if (r.pos==1)  rex = singularReductions.get(this, r);
                    if (rex==null) {
                        rex = r.rewrite(phase().getLocation());
                        if (r.pos==1) singularReductions.put(this, r, rex);
                    }
                    for(Node child : this.parents()) child.finish(r, rex, target, holder);
                } else {
                    for(Node child : this.parents()) child.reduce(r, pos-1, target, holder);
                }
                holder[pos] = old;
            }

            public void finish(Position r, Forest result, Phase<Tok> target, Forest[] holder) {
                Parser.Table<Tok>.State<Tok> state0 = state.gotoSetNonTerminals.get(r.owner());
                if (result==null) throw new Error();
                if (state0!=null)
                    target.newNode(this, result, state0, r.pos<=0, r);
            }

            public void performEmptyReductions() { state.invokeReductions(token, this, null, null); }

            private Node(Node parent, Forest pending, State state) {
                this.state = state;
                this.holder().merge(pending);
                Phase start = parent==null ? null : parent.phase();
                if (parent != null) parents().add(parent, true);
                if (Phase.this.hash.get(state, start) != null) throw new Error("severe problem!");
                Phase.this.hash.put(state, start, this);
            }
            public int toInt() { return idx; }
            private final int idx = node_idx++;
        }
        private int node_idx = 0;

        public int toInt() { return pos+1; }
        public int size() { return hash==null ? 0 : hash.size(); }
    }

}
